@@ -35,7 +35,7 @@ static char **filelist = NULL;
 static size_t filelist_len = 0;
 	/* The number of files in the list. */
 static int width = 0;
-	/* The number of files that we can display per line. */
+	/* The number of files that we can display per screen row. */
 static int longest = 0;
 	/* The number of columns in the longest filename in the list. */
 static size_t selected = 0;
@@ -385,7 +385,7 @@ char *do_browse_from(const char *inpath)
  * set filelist_len to the number of files in that list, set longest to
  * the width in columns of the longest filename in that list (between 15
  * and COLS), and set width to the number of files that we can display
- * per line.  And sort the list too. */
+ * per screen row.  And sort the list too. */
 void read_the_list(const char *path, DIR *dir)
 {
     const struct dirent *nextdir;
@@ -482,15 +482,15 @@ functionptrtype parse_browser_input(int *kbinput)
     return func_from_key(kbinput);
 }
 
-/* Set width to the number of files that we can display per line, if
- * necessary, and display the list of files. */
+/* Set width to the number of files that we can display per screen row,
+ * if necessary, and display the list of files. */
 void browser_refresh(void)
 {
     size_t i;
-    int line = 0, col = 0;
-	/* The current line and column while the list is getting displayed. */
-    int the_line = 0, the_column = 0;
-	/* The line and column of the selected item. */
+    int row = 0, col = 0;
+	/* The current row and column while the list is getting displayed. */
+    int the_row = 0, the_column = 0;
+	/* The row and column of the selected item. */
     char *info;
 	/* The additional information that we'll display about a file. */
 
@@ -501,7 +501,7 @@ void browser_refresh(void)
 
     i = selected - selected % (editwinrows * width);
 
-    for (; i < filelist_len && line < editwinrows; i++) {
+    for (; i < filelist_len && row < editwinrows; i++) {
 	struct stat st;
 	const char *thename = tail(filelist[i]);
 		/* The filename we display, minus the path. */
@@ -524,16 +524,16 @@ void browser_refresh(void)
 	 * remember its location to be able to place the cursor on it. */
 	if (i == selected) {
 	    wattron(edit, hilite_attribute);
-	    the_line = line;
+	    the_row = row;
 	    the_column = col;
 	}
 
-	blank_line(edit, line, col, longest);
+	blank_row(edit, row, col, longest);
 
 	/* If the name is too long, we display something like "...ename". */
 	if (dots)
-	    mvwaddstr(edit, line, col, "...");
-	mvwaddstr(edit, line, dots ? col + 3 : col, disp);
+	    mvwaddstr(edit, row, col, "...");
+	mvwaddstr(edit, row, dots ? col + 3 : col, disp);
 
 	free(disp);
 
@@ -591,7 +591,7 @@ void browser_refresh(void)
 	    infolen = infomaxlen;
 	}
 
-	mvwaddstr(edit, line, col - infolen, info);
+	mvwaddstr(edit, row, col - infolen, info);
 
 	/* If this is the selected item, finish its highlighting. */
 	if (i == selected)
@@ -602,17 +602,17 @@ void browser_refresh(void)
 	/* Add some space between the columns. */
 	col += 2;
 
-	/* If the next entry isn't going to fit on the current line,
-	 * move to the next line. */
+	/* If the next entry isn't going to fit on the current row,
+	 * move to the next row. */
 	if (col > COLS - longest) {
-	    line++;
+	    row++;
 	    col = 0;
 	}
     }
 
     /* If requested, put the cursor on the selected item and switch it on. */
     if (ISSET(SHOW_CURSOR)) {
-	wmove(edit, the_line, the_column);
+	wmove(edit, the_row, the_column);
 	curs_set(1);
     }
 
