@@ -52,27 +52,55 @@ void get_homedir(void)
     }
 }
 
-/* Read a ssize_t from str, and store it in *val (if val is not NULL).
- * On error, we return FALSE and don't change *val.  Otherwise, we
- * return TRUE. */
-bool parse_num(const char *str, ssize_t *val)
+#ifdef ENABLE_LINENUMBERS
+/* Return the number of digits that the given integer n takes up. */
+int digits(ssize_t n)
+{
+    if (n < 100000) {
+        if (n < 1000) {
+            if (n < 100)
+                return 2;
+            else
+                return 3;
+        } else {
+            if (n < 10000)
+                return 4;
+            else
+                return 5;
+        }
+    } else {
+        if (n < 10000000) {
+            if (n < 1000000)
+                return 6;
+            else
+                return 7;
+        }
+        else {
+            if (n < 100000000)
+                return 8;
+            else
+                return 9;
+        }
+    }
+}
+#endif
+
+/* Read an integer from str.  If it parses okay, store it in *result
+ * and return TRUE; otherwise, return FALSE. */
+bool parse_num(const char *str, ssize_t *result)
 {
     char *first_error;
-    ssize_t j;
+    ssize_t value;
 
-    assert(str != NULL);
-
-    /* The manual page for strtol() says this is required, and
-     * it looks like it is! */
+    /* The manual page for strtol() says this is required. */
     errno = 0;
 
-    j = (ssize_t)strtol(str, &first_error, 10);
+    value = (ssize_t)strtol(str, &first_error, 10);
 
     if (errno == ERANGE || *str == '\0' || *first_error != '\0')
 	return FALSE;
 
-    if (val != NULL)
-	*val = j;
+    *result = value;
 
     return TRUE;
 }
@@ -93,26 +121,24 @@ bool parse_line_column(const char *str, ssize_t *line, ssize_t *column)
     if (comma == NULL)
 	return parse_num(str, line);
 
-    if (!parse_num(comma + 1, column))
-	return FALSE;
+    retval = parse_num(comma + 1, column);
 
     if (comma == str)
-	return TRUE;
+	return retval;
 
     firstpart = mallocstrcpy(NULL, str);
     firstpart[comma - str] = '\0';
 
-    retval = parse_num(firstpart, line);
+    retval = parse_num(firstpart, line) && retval;
+
     free(firstpart);
 
     return retval;
 }
 
-/* Fix the memory allocation for a string. */
-void align(char **str)
+/* Reduce the memory allocation of a string to what is needed. */
+void snuggly_fit(char **str)
 {
-    assert(str != NULL);
-
     if (*str != NULL)
 	*str = charealloc(*str, strlen(*str) + 1);
 }
@@ -120,8 +146,6 @@ void align(char **str)
 /* Null a string at a certain index and align it. */
 void null_at(char **data, size_t index)
 {
-    assert(data != NULL);
-
     *data = charealloc(*data, index + 1);
     (*data)[index] = '\0';
 }
@@ -130,8 +154,6 @@ void null_at(char **data, size_t index)
  * normally have newlines in it, so encode its nulls as newlines. */
 void unsunder(char *str, size_t true_len)
 {
-    assert(str != NULL);
-
     for (; true_len > 0; true_len--, str++) {
 	if (*str == '\0')
 	    *str = '\n';
@@ -142,8 +164,6 @@ void unsunder(char *str, size_t true_len)
  * normally have newlines in it, so decode its newlines as nulls. */
 void sunder(char *str)
 {
-    assert(str != NULL);
-
     for (; *str != '\0'; str++) {
 	if (*str == '\n')
 	    *str = '\0';
@@ -280,11 +300,8 @@ const char *fixbounds(const char *r)
  * a separate word?  That is: is it not part of a longer word?*/
 bool is_separate_word(size_t position, size_t length, const char *buf)
 {
-    char *before = charalloc(mb_cur_max()), *after = charalloc(mb_cur_max());
+    char before[mb_cur_max()], after[mb_cur_max()];
     size_t word_end = position + length;
-    bool retval;
-
-    assert(buf != NULL && position < strlen(buf) && position + length <= strlen(buf));
 
     /* Get the characters before and after the word, if any. */
     parse_mbchar(buf + move_mbleft(buf, position), before, NULL);
@@ -293,77 +310,81 @@ bool is_separate_word(size_t position, size_t length, const char *buf)
     /* If the word starts at the beginning of the line OR the character before
      * the word isn't a letter, and if the word ends at the end of the line OR
      * the character after the word isn't a letter, we have a whole word. */
-    retval = (position == 0 || !is_alpha_mbchar(before)) &&
-		(word_end == strlen(buf) || !is_alpha_mbchar(after));
-
-    free(before);
-    free(after);
-
-    return retval;
+    return ((position == 0 || !is_alpha_mbchar(before)) &&
+		(buf[word_end] == '\0' || !is_alpha_mbchar(after)));
 }
 #endif /* !DISABLE_SPELLER */
 
-/* If we are searching backwards, we will find the last match that
- * starts no later than start.  Otherwise we find the first match
- * starting no earlier than start.  If we are doing a regexp search, we
- * fill in the global variable regmatches with at most 9 subexpression
- * matches.  Also, all .rm_so elements are relative to the start of the
- * whole match, so regmatches[0].rm_so == 0. */
+/* Return the position of the needle in the haystack, or NULL if not found.
+ * When searching backwards, we will find the last match that starts no later
+ * than the given start; otherwise, we find the first match starting no earlier
+ * than start.  If we are doing a regexp search, and we find a match, we fill
+ * in the global variable regmatches with at most 9 subexpression matches. */
 const char *strstrwrapper(const char *haystack, const char *needle,
 	const char *start)
 {
-    /* start can be 1 character before the start or after the end of the
-     * line.  In either case, we just say no match was found. */
-    if ((start > haystack && *(start - 1) == '\0') || start < haystack)
-	return NULL;
-
-    assert(haystack != NULL && needle != NULL && start != NULL);
-
 #ifdef HAVE_REGEX_H
     if (ISSET(USE_REGEXP)) {
-#ifndef NANO_TINY
 	if (ISSET(BACKWARDS_SEARCH)) {
-	    if (regexec(&search_regexp, haystack, 1, regmatches,
-		0) == 0 && haystack + regmatches[0].rm_so <= start) {
-		const char *retval = haystack + regmatches[0].rm_so;
+	    size_t last_find, ceiling, far_end;
+	    size_t floor = 0, next_rung = 0;
+		/* The start of the search range, and the next start. */
 
-		/* Search forward until there are no more matches. */
-		while (regexec(&search_regexp, retval + 1, 1,
-			regmatches, REG_NOTBOL) == 0 &&
-			retval + regmatches[0].rm_so + 1 <= start)
-		    retval += regmatches[0].rm_so + 1;
-		/* Finally, put the subexpression matches in global
-		 * variable regmatches.  The REG_NOTBOL flag doesn't
-		 * matter now. */
-		regexec(&search_regexp, retval, 10, regmatches, 0);
-		return retval;
+	    if (regexec(&search_regexp, haystack, 1, regmatches, 0) != 0)
+		return NULL;
+
+	    far_end = strlen(haystack);
+	    ceiling = start - haystack;
+	    last_find = regmatches[0].rm_so;
+
+	    /* A result beyond the search range also means: no match. */
+	    if (last_find > ceiling)
+		return NULL;
+
+	    /* Move the start-of-range forward until there is no more match;
+	     * then the last match found is the first match backwards. */
+	    while (regmatches[0].rm_so <= ceiling) {
+		floor = next_rung;
+		last_find = regmatches[0].rm_so;
+		/* If this is the last possible match, don't try to advance. */
+		if (last_find == ceiling)
+		    break;
+		next_rung = move_mbright(haystack, last_find);
+		regmatches[0].rm_so = next_rung;
+		regmatches[0].rm_eo = far_end;
+		if (regexec(&search_regexp, haystack, 1, regmatches,
+					REG_STARTEND) != 0)
+		    break;
 	    }
-	} else
-#endif /* !NANO_TINY */
-	if (regexec(&search_regexp, start, 10, regmatches,
-		(start > haystack) ? REG_NOTBOL : 0) == 0) {
-	    const char *retval = start + regmatches[0].rm_so;
 
-	    regexec(&search_regexp, retval, 10, regmatches, 0);
-	    return retval;
+	    /* Find the last match again, to get possible submatches. */
+	    regmatches[0].rm_so = floor;
+	    regmatches[0].rm_eo = far_end;
+	    if (regexec(&search_regexp, haystack, 10, regmatches,
+					REG_STARTEND) != 0)
+		statusline(ALERT, "BAD: failed to refind the match!");
+
+	    return haystack + regmatches[0].rm_so;
 	}
-	return NULL;
+
+	/* Do a forward regex search from the starting point. */
+	regmatches[0].rm_so = start - haystack;
+	regmatches[0].rm_eo = strlen(haystack);
+	if (regexec(&search_regexp, haystack, 10, regmatches,
+					REG_STARTEND) != 0)
+	    return NULL;
+	else
+	    return haystack + regmatches[0].rm_so;
     }
 #endif /* HAVE_REGEX_H */
-#if !defined(NANO_TINY) || !defined(DISABLE_SPELLER)
     if (ISSET(CASE_SENSITIVE)) {
-#ifndef NANO_TINY
 	if (ISSET(BACKWARDS_SEARCH))
 	    return revstrstr(haystack, needle, start);
 	else
-#endif
 	    return strstr(start, needle);
-    }
-#endif /* !DISABLE_SPELLER || !NANO_TINY */
-#ifndef NANO_TINY
-    else if (ISSET(BACKWARDS_SEARCH))
+    } else if (ISSET(BACKWARDS_SEARCH))
 	return mbrevstrcasestr(haystack, needle, start);
-#endif
+
     return mbstrcasestr(start, needle);
 }
 
@@ -439,20 +460,18 @@ char *free_and_assign(char *dest, char *src)
  * get_page_start(column) < COLS). */
 size_t get_page_start(size_t column)
 {
-    if (column == 0 || column < COLS - 1)
+    if (column == 0 || column < editwincols - 1)
 	return 0;
-    else if (COLS > 8)
-	return column - 7 - (column - 7) % (COLS - 8);
+    else if (editwincols > 8)
+	return column - 7 - (column - 7) % (editwincols - 8);
     else
-	return column - (COLS - 2);
+	return column - (editwincols - 2);
 }
 
 /* Return the placewewant associated with current_x, i.e. the zero-based
  * column position of the cursor. */
 size_t xplustabs(void)
 {
-    assert(openfile->current != NULL);
-
     return strnlenpt(openfile->current->data, openfile->current_x);
 }
 
@@ -465,8 +484,6 @@ size_t actual_x(const char *text, size_t column)
     size_t width = 0;
 	/* The screen display width to text[index], in columns. */
 
-    assert(text != NULL);
-
     while (*text != '\0') {
 	int charlen = parse_mbchar(text, NULL, &width);
 
@@ -490,8 +507,6 @@ size_t strnlenpt(const char *text, size_t maxlen)
     if (maxlen == 0)
 	return 0;
 
-    assert(text != NULL);
-
     while (*text != '\0') {
 	int charlen = parse_mbchar(text, NULL, &width);
 
@@ -515,14 +530,8 @@ size_t strlenpt(const char *text)
 /* Append a new magicline to filebot. */
 void new_magicline(void)
 {
-    openfile->filebot->next = (filestruct *)nmalloc(sizeof(filestruct));
+    openfile->filebot->next = make_new_node(openfile->filebot);
     openfile->filebot->next->data = mallocstrcpy(NULL, "");
-    openfile->filebot->next->prev = openfile->filebot;
-    openfile->filebot->next->next = NULL;
-    openfile->filebot->next->lineno = openfile->filebot->lineno + 1;
-#ifndef DISABLE_COLOR
-    openfile->filebot->next->multidata = NULL;
-#endif
     openfile->filebot = openfile->filebot->next;
     openfile->totsize++;
 }
@@ -535,9 +544,6 @@ void remove_magicline(void)
 {
     if (openfile->filebot->data[0] == '\0' &&
 		openfile->filebot != openfile->fileage) {
-	assert(openfile->filebot != openfile->edittop &&
-		openfile->filebot != openfile->current);
-
 	openfile->filebot = openfile->filebot->prev;
 	free_filestruct(openfile->filebot->next);
 	openfile->filebot->next = NULL;
@@ -553,8 +559,6 @@ void remove_magicline(void)
 void mark_order(const filestruct **top, size_t *top_x, const filestruct
 	**bot, size_t *bot_x, bool *right_side_up)
 {
-    assert(top != NULL && top_x != NULL && bot != NULL && bot_x != NULL);
-
     if ((openfile->current->lineno == openfile->mark_begin->lineno &&
 		openfile->current_x > openfile->mark_begin_x) ||
 		openfile->current->lineno > openfile->mark_begin->lineno) {
@@ -573,37 +577,6 @@ void mark_order(const filestruct **top, size_t *top_x, const filestruct
 	    *right_side_up = FALSE;
     }
 }
-#endif /* !NANO_TINY */
-
-/* Calculate the number of characters between begin and end, and return
- * it. */
-size_t get_totsize(const filestruct *begin, const filestruct *end)
-{
-    size_t totsize = 0;
-    const filestruct *f;
-
-    /* Go through the lines from begin to end->prev, if we can. */
-    for (f = begin; f != end && f != NULL; f = f->next) {
-	/* Count the number of characters on this line. */
-	totsize += mbstrlen(f->data);
-
-	/* Count the newline if we have one. */
-	if (f->next != NULL)
-	    totsize++;
-    }
-
-    /* Go through the line at end, if we can. */
-    if (f != NULL) {
-	/* Count the number of characters on this line. */
-	totsize += mbstrlen(f->data);
-
-	/* Count the newline if we have one. */
-	if (f->next != NULL)
-	    totsize++;
-    }
-
-    return totsize;
-}
 
 /* Given a line number, return a pointer to the corresponding struct. */
 filestruct *fsfromline(ssize_t lineno)
@@ -618,13 +591,32 @@ filestruct *fsfromline(ssize_t lineno)
 	    f = f->next;
 
     if (f->lineno != lineno) {
-	statusline(ALERT, _("Internal error: can't match line %d.  "
-			"Please save your work."), lineno);
+	statusline(ALERT, _("Internal error: can't match line %ld.  "
+			"Please save your work."), (long)lineno);
 	return NULL;
     }
 
     return f;
 }
+#endif /* !NANO_TINY */
+
+/* Count the number of characters from begin to end, and return it. */
+size_t get_totsize(const filestruct *begin, const filestruct *end)
+{
+    const filestruct *line;
+    size_t totsize = 0;
+
+    /* Sum the number of characters (plus a newline) in each line. */
+    for (line = begin; line != end->next; line = line->next)
+	totsize += mbstrlen(line->data) + 1;
+
+    /* The last line of a file doesn't have a newline -- otherwise it
+     * wouldn't be the last line -- so subtract 1 when at EOF. */
+    if (line == NULL)
+	totsize--;
+
+    return totsize;
+}
 
 #ifdef DEBUG
 /* Dump the filestruct inptr to stderr. */
